"""
Helps to access the unit catalog and find physical dimensions.
"""

from pandas import DataFrame

import odsbox.proto.ods_pb2 as ods
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .con_i import ConI
from odsbox.datamatrices_to_pandas import to_pandas


def query_physical_dimensions(
    con_i: "ConI",
    length: int = 0,
    mass: int = 0,
    time: int = 0,
    current: int = 0,
    temperature: int = 0,
    molar_amount: int = 0,
    luminous_intensity: int = 0,
    **kwargs,
) -> DataFrame:
    """
    Search for a physical dimension by its SI exponents.
    """
    physical_dimension_entity = con_i.mc.entity_by_base_name("AoPhysicalDimension")

    select = ods.SelectStatement()
    select.columns.add(aid=physical_dimension_entity.aid, attribute="*")
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "length_exp").name
    ci.condition.long_array.values.append(length)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "mass_exp").name
    ci.condition.long_array.values.append(mass)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "time_exp").name
    ci.condition.long_array.values.append(time)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "current_exp").name
    ci.condition.long_array.values.append(current)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "temperature_exp").name
    ci.condition.long_array.values.append(temperature)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "molar_amount_exp").name
    ci.condition.long_array.values.append(molar_amount)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "luminous_intensity_exp").name
    ci.condition.long_array.values.append(luminous_intensity)
    return to_pandas(con_i.data_read(select), **kwargs)


def query_units(
    con_i: "ConI",
    length: int = 0,
    mass: int = 0,
    time: int = 0,
    current: int = 0,
    temperature: int = 0,
    molar_amount: int = 0,
    luminous_intensity: int = 0,
    **kwargs,
) -> DataFrame:
    """
    Search for a unit by its SI exponents.
    """
    unit_entity = con_i.mc.entity_by_base_name("AoUnit")
    physical_dimension_entity = con_i.mc.entity_by_base_name("AoPhysicalDimension")

    select = ods.SelectStatement()  # pylint: disable=E1101
    select.columns.add(aid=unit_entity.aid, attribute="*")
    select.columns.add(aid=physical_dimension_entity.aid, attribute="Name")
    select.joins.add(
        aid_from=unit_entity.aid,
        aid_to=physical_dimension_entity.aid,
        relation=con_i.mc.relation_by_base_name(unit_entity, "phys_dimension").name,
    )
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "length_exp").name
    ci.condition.long_array.values.append(length)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "mass_exp").name
    ci.condition.long_array.values.append(mass)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "time_exp").name
    ci.condition.long_array.values.append(time)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "current_exp").name
    ci.condition.long_array.values.append(current)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "temperature_exp").name
    ci.condition.long_array.values.append(temperature)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "molar_amount_exp").name
    ci.condition.long_array.values.append(molar_amount)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "luminous_intensity_exp").name
    ci.condition.long_array.values.append(luminous_intensity)
    return to_pandas(
        con_i.data_read(select),
        **kwargs,
    )


def query_quantity(
    con_i: "ConI",
    length: int = 0,
    mass: int = 0,
    time: int = 0,
    current: int = 0,
    temperature: int = 0,
    molar_amount: int = 0,
    luminous_intensity: int = 0,
    **kwargs,
) -> DataFrame:
    """
    Search for a quantity by its SI exponents.
    """
    unit_entity = con_i.mc.entity_by_base_name("AoUnit")
    physical_dimension_entity = con_i.mc.entity_by_base_name("AoPhysicalDimension")
    quantity_entity = con_i.mc.entity_by_base_name("AoQuantity")

    select = ods.SelectStatement()
    select.columns.add(aid=quantity_entity.aid, attribute="*")
    select.columns.add(aid=unit_entity.aid, attribute="Name")
    select.columns.add(aid=physical_dimension_entity.aid, attribute="Name")
    select.joins.add(
        aid_from=quantity_entity.aid,
        aid_to=unit_entity.aid,
        relation=con_i.mc.relation_by_base_name(quantity_entity, "default_unit").name,
    )
    select.joins.add(
        aid_from=unit_entity.aid,
        aid_to=physical_dimension_entity.aid,
        relation=con_i.mc.relation_by_base_name(unit_entity, "phys_dimension").name,
    )
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "length_exp").name
    ci.condition.long_array.values.append(length)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "mass_exp").name
    ci.condition.long_array.values.append(mass)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "time_exp").name
    ci.condition.long_array.values.append(time)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "current_exp").name
    ci.condition.long_array.values.append(current)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "temperature_exp").name
    ci.condition.long_array.values.append(temperature)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "molar_amount_exp").name
    ci.condition.long_array.values.append(molar_amount)
    ci = select.where.add()
    ci.condition.aid = physical_dimension_entity.aid
    ci.condition.attribute = con_i.mc.attribute_by_base_name(physical_dimension_entity, "luminous_intensity_exp").name
    ci.condition.long_array.values.append(luminous_intensity)
    return to_pandas(con_i.data_read(select), **kwargs)



"""Helper class for ASAM ODS Security configuration"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .con_i import ConI
import odsbox.proto.ods_security_pb2 as ods_security

from enum import IntFlag


class Security:
    """
    This class offers the ASAM ODS server security API
    """

    def __init__(self, con_i: "ConI"):
        self.__con_i = con_i

    def security_read(self, security_read_request: ods_security.SecurityReadRequest) -> ods_security.SecurityInfo:
        """
        This method reads the security configuration from the ASAM ODS server.
        :param security_read_request: The security read request
        :raises requests.HTTPError: If fails.
        :return: The security information
        """
        response = self.__con_i.ods_post_request("security-read", security_read_request)
        security_info = ods_security.SecurityInfo()
        security_info.ParseFromString(response.content)
        return security_info

    def security_update(self, security_write_request: ods_security.SecurityWriteRequest) -> None:
        """
        This method updates the security configuration on the ASAM ODS server.
        :param security_write_request: The security write request
        :raises requests.HTTPError: If fails.
        """
        self.__con_i.ods_post_request("security-update", security_write_request)

    def initial_rights(self, security_write_request: ods_security.SecurityWriteRequest) -> None:
        """
        This method sets the initial rights for newly created instances.
        :param security_write_request: The security write request
        :raises requests.HTTPError: If fails.
        """
        self.__con_i.ods_post_request("initial-rights", security_write_request)

    class Level(IntFlag):
        """
        This class defines the security levels for the ASAM ODS server.
        The levels are defined as bit flags found in the security_level attribute
        of the entities.
        """

        ELEMENT = 1  # Bit 0
        INSTANCE = 2  # Bit 1
        ATTRIBUTE = 4  # Bit 2

    class Right(IntFlag):
        """
        This class defines the security rights for the ASAM ODS server.
        The rights are defined as bit flags found in the security api.
        In the API they are used as integer values.
        """

        READ = 1  # Bit 0
        UPDATE = 2  # Bit 1
        INSERT = 4  # Bit 2
        DELETE = 8  # Bit 3
        GRANT = 16  # Bit 4



"""datetime is represented as string in ASAM ODS formatted using YYYYMMDDHHMMSSFFF. Here you find some helpers."""

from __future__ import annotations

import pandas as pd


def __normalize_datetime_string(asam_time):
    asam_time_len = len(asam_time)
    if asam_time_len < 8:
        raise SyntaxError("Time value mus at least contain year, month, day")
    if asam_time_len <= 14:
        return asam_time.ljust(14, "0")
    if asam_time_len <= 20:
        return asam_time.ljust(20, "0")
    if asam_time_len <= 23:
        return asam_time.ljust(23, "0")
    return asam_time


def to_pd_timestamp(asam_time: str) -> pd.Timestamp:
    """
    Convert ASAM ODS datetime string to pandas Timestamp.

    :param str asam_time: ASAM ODS datetime string to be converted. formatted like `YYYYMMDDHHMMSSFFF`.
                          It must at least contain `YYYYMMDD`.
    :raises requests.SyntaxError: If content is invalid.
    :return pd.Timestamp: Corresponding pandas Timestamp value. For empty string `pd.NaT` is returned.
    """
    if asam_time is None or "" == asam_time:
        return pd.NaT  # type: ignore

    asam_time_normalized = __normalize_datetime_string(asam_time)
    asam_time_len = len(asam_time_normalized)

    return pd.Timestamp(
        year=int(asam_time_normalized[0:4]) if asam_time_len >= 4 else 0,
        month=int(asam_time_normalized[4:6]) if asam_time_len >= 6 else 1,
        day=int(asam_time_normalized[6:8]) if asam_time_len >= 8 else 1,
        hour=int(asam_time_normalized[8:10]) if asam_time_len >= 10 else 0,
        minute=int(asam_time_normalized[10:12]) if asam_time_len >= 12 else 0,
        second=int(asam_time_normalized[12:14]) if asam_time_len >= 14 else 0,
        microsecond=int(asam_time_normalized[14:20]) if asam_time_len >= 20 else 0,
        nanosecond=int(asam_time_normalized[20:23]) if asam_time_len >= 23 else 0,
    )


def from_pd_timestamp(timestamp: pd.Timestamp, length: int = 17) -> str:
    """
    Convert a pandas Timestamp to a string formatted as asamtime (`YYYYMMDDHHMMSSFFF`).

    :param pd.Timestamp timestamp: The pandas Timestamp to convert. The timezone
                                   information given in timestamp is ignored.
    :param int length: The desired length of the output string. The final string will
                       be truncated to the specified length. The maximum is 23 including
                       nanoseconds.
    :return str: The asam time representation of the timestamp. For `None` or `pd.NaT`
                 an empty string is returned.
    """
    if timestamp is None or pd.isna(timestamp):
        return ""

    asam_time_str = timestamp.strftime("%Y%m%d%H%M%S%f")
    if length > 20:
        asam_time_str += f"{timestamp.nanosecond:03d}"
    return asam_time_str[: min(length, len(asam_time_str))]
